# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
#
# invenio-circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation JSON Resolver module."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019 CERN.
# Copyright (C) 2018-2019 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests for loan search class."""

from elasticsearch import VERSION as ES_VERSION

from invenio_circulation.api import Loan
from invenio_circulation.search.api import search_by_patron_item_or_document, \
    search_by_patron_pid, search_by_pid


def _assert_total(total, expected):
    """Assert total (ES6 compatibility)."""
    if ES_VERSION[0] >= 7:
        assert total.value == expected
    else:
        assert total == expected


def test_search_loans_by_pid(indexed_loans):
    """Test retrieve loan list belonging to an item."""
    loans = list(search_by_pid(item_pid="item_pending_1").scan())
    assert len(loans) == 1
    loan = Loan.get_record_by_pid(loans[0]["pid"])
    assert loan.get("item_pid") == "item_pending_1"


def test_search_loans_by_pid_filtering_states(indexed_loans):
    """Test retrieve loan list belonging to an item filtering states."""
    search = search_by_pid(item_pid="item_multiple_pending_on_loan_7",
                           filter_states=["PENDING", "ITEM_ON_LOAN"])
    search_result = search.execute()
    _assert_total(search_result.hits.total, 3)


def test_search_loans_by_pid_excluding_states(indexed_loans):
    """Test retrieve loan list belonging to an item excluding states."""
    search_result = search_by_pid(item_pid="item_multiple_pending_on_loan_7",
                                  exclude_states=["ITEM_ON_LOAN"]).execute()
    _assert_total(search_result.hits.total, 2)


def test_search_loans_by_patron_pid(indexed_loans):
    """Test retrieve loan list belonging to a patron."""
    search_result = search_by_patron_pid("1").execute()
    _assert_total(search_result.hits.total, 8)

    search_result = search_by_patron_pid("2").execute()
    _assert_total(search_result.hits.total, 3)

    search_result = search_by_patron_pid("3").execute()
    _assert_total(search_result.hits.total, 1)


def test_search_loans_by_patron_and_item_or_document(indexed_loans):
    """Test retrieve loan list by patron and items."""
    search_result = search_by_patron_item_or_document(
        patron_pid="1",
        item_pid="item_returned_3").execute()
    _assert_total(search_result.hits.total, 1)

    search_result = search_by_patron_item_or_document(
        patron_pid="1",
        item_pid="not_existing").execute()
    _assert_total(search_result.hits.total, 0)

    search_result = search_by_patron_item_or_document(
        patron_pid="999999",
        item_pid="item_returned_3").execute()
    _assert_total(search_result.hits.total, 0)

    search_result = search_by_patron_item_or_document(
        patron_pid="1",
        document_pid="document_pid").execute()
    _assert_total(search_result.hits.total, 8)

    search_result = search_by_patron_item_or_document(
        patron_pid="1",
        document_pid="not_existing").execute()
    _assert_total(search_result.hits.total, 0)

    search_result = search_by_patron_item_or_document(
        patron_pid="999999",
        document_pid="document_returned_1").execute()
    _assert_total(search_result.hits.total, 0)


def test_search_loans_by_patron_and_item_or_document_filtering_states(
     indexed_loans):
    """Test retrieve loan list by patron and items filtering states."""
    search = search_by_patron_item_or_document(patron_pid="1",
                                               item_pid="item_returned_3",
                                               filter_states=['ITEM_RETURNED'])
    search_result = search.execute()
    _assert_total(search_result.hits.total, 1)

    search = search_by_patron_item_or_document(patron_pid="1",
                                               item_pid="item_returned_3",
                                               filter_states=['ITEM_AT_DESK'])
    search_result = search.execute()
    _assert_total(search_result.hits.total, 0)

    search = search_by_patron_item_or_document(patron_pid="1",
                                               document_pid="document_pid",
                                               filter_states=['ITEM_RETURNED'])
    search_result = search.execute()
    _assert_total(search_result.hits.total, 2)

    search = search_by_patron_item_or_document(patron_pid="2",
                                               document_pid="document_pid",
                                               filter_states=['ITEM_RETURNED'])
    search_result = search.execute()
    _assert_total(search_result.hits.total, 1)

    search = search_by_patron_item_or_document(patron_pid="1",
                                               document_pid="document_pid",
                                               filter_states=['ITEM_AT_DESK'])
    search_result = search.execute()
    _assert_total(search_result.hits.total, 1)

    search = search_by_patron_item_or_document(patron_pid="1",
                                               document_pid="document_pid",
                                               filter_states=['PENDING'])
    search_result = search.execute()
    _assert_total(search_result.hits.total, 3)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019 CERN.
# Copyright (C) 2018-2019 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Circulation custom transitions."""

from flask import current_app
from invenio_db import db

from invenio_circulation.proxies import current_circulation

from ..api import can_be_requested, get_available_item_by_doc_pid, \
    get_document_pid_by_item_pid, get_pending_loans_by_doc_pid
from ..errors import ItemDoNotMatchError, ItemNotAvailableError, \
    LoanMaxExtensionError, RecordCannotBeRequestedError, \
    TransitionConditionsFailedError, TransitionConstraintsViolationError
from ..transitions.base import Transition
from ..transitions.conditions import is_same_location


def _ensure_valid_loan_duration(loan):
    """Validate start and end dates for a loan."""
    loan.setdefault('start_date', loan['transaction_date'])

    if not loan.get('end_date'):
        get_loan_duration = current_app.config['CIRCULATION_POLICIES'][
            'checkout']['duration_default']
        duration = get_loan_duration(loan)
        loan['end_date'] = loan['start_date'] + duration

    is_duration_valid = current_app.config['CIRCULATION_POLICIES']['checkout'][
        'duration_validate']
    if not is_duration_valid(loan):
        msg = "The loan duration from '{0}' to '{1}' is not valid.".format(
            loan['start_date'].isoformat(),
            loan['end_date'].isoformat()
        )
        raise TransitionConstraintsViolationError(description=msg)


def _ensure_item_attached_to_loan(loan):
    """Validate that an item is attached to a loan."""
    if not loan.get('item_pid'):
        msg = "No item assigned to loan '{0}'.".format(loan.id)
        raise TransitionConditionsFailedError(description=msg)


def ensure_same_item(f):
    """Validate that the item PID exists and cannot be changed."""
    def inner(self, loan, **kwargs):
        new_item_pid = kwargs.get('item_pid')

        if new_item_pid:
            if not current_app \
                    .config['CIRCULATION_ITEM_EXISTS'](new_item_pid):
                msg = "Item '{0}' not found in the system"\
                    .format(new_item_pid)
                raise ItemNotAvailableError(description=msg)

            if loan.get('item_pid') \
               and new_item_pid != loan['item_pid']:
                msg = "Cannot change item '{}' while performing an action on "\
                      "this loan".format(new_item_pid)
                raise ItemDoNotMatchError(description=msg)

        return f(self, loan, **kwargs)
    return inner


def _update_document_pending_request_for_item(item_pid, **kwargs):
    """Update pending loans on a Document with no Item attached yet."""
    document_pid = get_document_pid_by_item_pid(item_pid)
    for pending_loan in get_pending_loans_by_doc_pid(document_pid):
        pending_loan['item_pid'] = item_pid
        pending_loan.commit()
        db.session.commit()
        current_circulation.loan_indexer.index(pending_loan)


def _ensure_valid_extension(loan):
    """Validate end dates for a extended loan."""
    extension_count = loan.get('extension_count', 0)
    extension_count += 1

    get_extension_max_count_func = current_app.config['CIRCULATION_POLICIES'][
        'extension']['max_count']
    extension_max_count = get_extension_max_count_func(loan)
    if extension_count > extension_max_count:
        raise LoanMaxExtensionError(
            loan_pid=loan["pid"],
            extension_count=extension_max_count
        )
    loan['extension_count'] = extension_count

    get_extension_duration_func = current_app.config['CIRCULATION_POLICIES'][
        'extension']['duration_default']
    duration = get_extension_duration_func(loan)

    should_extend_from_end_date = current_app.config[
        'CIRCULATION_POLICIES']['extension']['from_end_date']
    if not should_extend_from_end_date:
        # extend from the transaction_date instead
        loan['end_date'] = loan['transaction_date']

    loan['end_date'] += duration


def _ensure_same_location(item_pid, location_pid, destination, error_msg):
    """Validate that item location is same as loan pickup location."""
    if not is_same_location(item_pid, location_pid):
        error_msg += "Transition to '{}' has failed.".format(destination)
        raise TransitionConditionsFailedError(description=error_msg)


def _ensure_not_same_location(item_pid, location_pid, destination, error_msg):
    """Validate that item location is not the same as loan pickup location."""
    if is_same_location(item_pid, location_pid):
        error_msg += "Transition to '{}' has failed.".format(destination)
        raise TransitionConditionsFailedError(description=error_msg)


def _get_item_location(item_pid):
    """Retrieve Item location based on PID."""
    return current_app.config[
        'CIRCULATION_ITEM_LOCATION_RETRIEVER'](item_pid)


class ToItemOnLoan(Transition):
    """Action to checkout."""

    def before(self, loan, **kwargs):
        """Validate checkout action."""
        super().before(loan, **kwargs)

        self.ensure_item_is_available_for_checkout(loan)

        if not kwargs.get("pickup_location_pid") \
           or "pickup_location_pid" not in loan:
            loan['pickup_location_pid'] = _get_item_location(loan['item_pid'])

        _ensure_valid_loan_duration(loan)

    def after(self, loan):
        """Attach Item after checkout."""
        if 'item' not in loan:
            loan.attach_item_ref()
        super().after(loan)


class ItemAtDeskToItemOnLoan(ToItemOnLoan):
    """Check-out action to perform a loan when item ready at desk."""

    def before(self, loan, **kwargs):
        """Validate checkout action."""
        super().before(loan, **kwargs)
        _ensure_valid_loan_duration(loan)


def check_request_on_document(f):
    """Decorator to check if the request is on document."""
    def inner(self, loan, **kwargs):
        document_pid = kwargs.get('document_pid')
        if document_pid and not kwargs.get('item_pid'):
            if not can_be_requested(loan):
                msg = "Cannot create a request for the document '{}'" \
                    .format(loan.get('document_pid'))
                raise RecordCannotBeRequestedError(description=msg)

            if self.assign_item:
                available_item_pid = get_available_item_by_doc_pid(
                    document_pid
                )
                if available_item_pid:
                    kwargs['item_pid'] = available_item_pid

        if kwargs.get("item_pid") \
           and not kwargs.get("pickup_location_pid"):
            # if no pickup location was specified in the request,
            # assign a default one
            kwargs['pickup_location_pid'] = _get_item_location(
                                                kwargs.get("item_pid"))

        return f(self, loan, **kwargs)
    return inner


class CreatedToPending(Transition):
    """Action to request to loan an item."""

    def __init__(self, src, dest, trigger='next', permission_factory=None,
                 **kwargs):
        """Constructor."""
        super().__init__(src, dest, trigger=trigger,
                         permission_factory=permission_factory, **kwargs)
        self.assign_item = kwargs.get('assign_item', True)

    @check_request_on_document
    def before(self, loan, **kwargs):
        """Set a default pickup location if not passed as param."""
        super().before(loan, **kwargs)

        if not can_be_requested(loan):
            msg = "Cannot create a request for the item '{}'" \
                .format(loan.get('item_pid'))
            raise RecordCannotBeRequestedError(description=msg)


class PendingToItemAtDesk(Transition):
    """Validate pending request to prepare the item at desk of its location."""

    def before(self, loan, **kwargs):
        """Validate if the item is for this location or should transit."""
        super().before(loan, **kwargs)

        # check if a request on document has no item attached
        _ensure_item_attached_to_loan(loan)
        _ensure_same_location(loan['item_pid'],
                              loan['pickup_location_pid'],
                              self.dest,
                              error_msg="Pickup is not at the same library. ")


class PendingToItemInTransitPickup(Transition):
    """Validate pending request to send the item to the pickup location."""

    def before(self, loan, **kwargs):
        """Validate if the item is for this location or should transit."""
        super().before(loan, **kwargs)

        # check if a request on document has no item attached
        _ensure_item_attached_to_loan(loan)
        _ensure_not_same_location(loan['item_pid'],
                                  loan['pickup_location_pid'],
                                  self.dest,
                                  error_msg="Pickup is at the same library. ")


class ItemOnLoanToItemOnLoan(Transition):
    """Extend action to perform a item loan extension."""

    @ensure_same_item
    def before(self, loan, **kwargs):
        """Validate extension action."""
        super().before(loan, **kwargs)
        _ensure_valid_extension(loan)


class ItemOnLoanToItemInTransitHouse(Transition):
    """Check-in action when returning an item not to its belonging location."""

    @ensure_same_item
    def before(self, loan, **kwargs):
        """Validate check-in action."""
        super().before(loan, **kwargs)

        _ensure_not_same_location(loan['item_pid'],
                                  loan['transaction_location_pid'],
                                  self.dest,
                                  error_msg="Item should be returned "
                                            "(already in house). ")

        # set end loan date as transaction date when completing loan
        loan['end_date'] = loan['transaction_date']


class ItemOnLoanToItemReturned(Transition):
    """Check-in action when returning an item to its belonging location."""

    def __init__(self, src, dest, trigger='next', permission_factory=None,
                 **kwargs):
        """Constructor."""
        super().__init__(src, dest, trigger=trigger,
                         permission_factory=permission_factory, **kwargs)
        self.assign_item = kwargs.get('assign_item', True)

    @ensure_same_item
    def before(self, loan, **kwargs):
        """Validate check-in action."""
        super().before(loan, **kwargs)

        _ensure_same_location(loan['item_pid'],
                              loan['transaction_location_pid'],
                              self.dest,
                              error_msg="Item should be in transit to house. ")

        # set end loan date as transaction date when completing loan
        loan['end_date'] = loan['transaction_date']

    def after(self, loan):
        """Check for pending requests on this item after check-in."""
        super().after(loan)
        if self.assign_item:
            _update_document_pending_request_for_item(loan['item_pid'])


class ItemInTransitHouseToItemReturned(Transition):
    """Check-in action when returning an item to its belonging location."""

    def __init__(self, src, dest, trigger='next', permission_factory=None,
                 **kwargs):
        """Constructor."""
        super().__init__(src, dest, trigger=trigger,
                         permission_factory=permission_factory, **kwargs)
        self.assign_item = kwargs.get('assign_item', True)

    @ensure_same_item
    def before(self, loan, **kwargs):
        """Validate check-in action."""
        super().before(loan, **kwargs)

        _ensure_same_location(loan['item_pid'],
                              loan['transaction_location_pid'],
                              self.dest,
                              error_msg="Item should be in transit to house. ")

    def after(self, loan):
        """Check for pending requests on this item after check-in."""
        super().after(loan)
        if self.assign_item:
            _update_document_pending_request_for_item(loan['item_pid'])


class ToCancelled(Transition):
    """When cancelling a loan, ensure that the item is not changed."""

    @ensure_same_item
    def before(self, loan, **kwargs):
        """Validate cancel action."""
        super().before(loan, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
# Copyright (C) 2019 RERO.
#
# invenio-circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation record loaders JSON schemas."""


import arrow
from flask_babelex import lazy_gettext as _
from invenio_records_rest.schemas import RecordMetadataSchemaJSONV1
from invenio_records_rest.schemas.fields import PersistentIdentifier
from marshmallow import ValidationError, fields


class DateTimeString(fields.DateTime):
    """Custom DateTime field to return a string representation."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        kwargs.setdefault('validate', self.validate_timezone)
        super().__init__(*args, **kwargs)

    def validate_timezone(self, value):
        """Validate that the passed timezone, if any, is UTC."""
        # strong validation of input datetime to require UTC timezone
        if arrow.get(value).isoformat() != \
                arrow.get(value).to('utc').isoformat():
            raise ValidationError(_("Not a valid ISO-8601 UTC datetime."))

    def deserialize(self, value, attr=None, data=None, **kwargs):
        """Validate ISO8601 datetime input but return the string value."""
        super().deserialize(value, attr, data, **kwargs)
        # return the original string value after marshmallow validation
        # because Invenio does not support Python datetime JSON serializer yet
        return value


class DateString(fields.Date):
    """Custom Date field to return a string representation."""

    def deserialize(self, value, attr=None, data=None, **kwargs):
        """Validate ISO8601 date input but return the string value."""
        super().deserialize(value, attr, data, **kwargs)
        # return the original string value after marshmallow validation
        # because Invenio does not support Python datetime JSON serializer yet
        return value


class LoanSchemaV1(RecordMetadataSchemaJSONV1):
    """Loan schema."""

    class Meta:
        """Meta attributes for the schema."""

        from marshmallow import EXCLUDE
        unknown = EXCLUDE

    def get_pid_field(self):
        """Return loan PID field name."""
        return "pid"

    cancel_reason = fields.Str()
    document_pid = fields.Str()
    end_date = DateString()
    extension_count = fields.Integer()
    item_pid = fields.Str()
    patron_pid = fields.Str(required=True)
    pid = PersistentIdentifier()
    pickup_location_pid = fields.Str()
    request_expire_date = DateString()
    request_start_date = DateString()
    start_date = DateString()
    transaction_date = DateTimeString()
    transaction_location_pid = fields.Str(required=True)
    transaction_user_pid = fields.Str(required=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019 CERN.
# Copyright (C) 2018-2019 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Circulation API."""

from elasticsearch import VERSION as ES_VERSION
from flask import current_app
from invenio_jsonschemas import current_jsonschemas
from invenio_pidstore.resolver import Resolver
from invenio_records.api import Record

from .errors import MissingRequiredParameterError, MultipleLoansOnItemError
from .pidstore.pids import CIRCULATION_LOAN_PID_TYPE
from .search.api import search_by_pid
from .utils import str2datetime


class Loan(Record):
    """Loan record class."""

    DATE_FIELDS = [
        "start_date",
        "end_date",
        "request_expire_date",
        "request_start_date",
    ]
    DATETIME_FIELDS = ["transaction_date"]

    _schema = "loans/loan-v1.0.0.json"

    def __init__(self, data, model=None):
        """Constructor."""
        self.item_ref_builder = current_app.config.get(
            "CIRCULATION_ITEM_REF_BUILDER"
        )
        self["state"] = current_app.config["CIRCULATION_LOAN_INITIAL_STATE"]
        super().__init__(data, model)

    @classmethod
    def create(cls, data, id_=None, **kwargs):
        """Create Loan record."""
        data["$schema"] = current_jsonschemas.path_to_url(cls._schema)
        ref_builder = current_app.config.get("CIRCULATION_ITEM_REF_BUILDER")
        data["item"] = ref_builder(data["pid"])
        item_pid = data.get("item_pid")
        if item_pid:
            data["document_pid"] = get_document_pid_by_item_pid(item_pid)

        return super().create(data, id_=id_, **kwargs)

    def date_fields2datetime(self):
        """Convert string datetime fields to Python datetime."""
        for field in self.DATE_FIELDS + self.DATETIME_FIELDS:
            if field in self:
                self[field] = str2datetime(self[field])

    def date_fields2str(self):
        """Convert Python datetime fields to string."""
        for field in self.DATE_FIELDS:
            if field in self:
                self[field] = self[field].date().isoformat()
        for field in self.DATETIME_FIELDS:
            if field in self:
                self[field] = self[field].isoformat()

    @classmethod
    def get_record_by_pid(cls, pid, with_deleted=False):
        """Get ils record by pid value."""
        resolver = Resolver(
            pid_type=CIRCULATION_LOAN_PID_TYPE,
            object_type="rec",
            getter=cls.get_record,
        )
        persistent_identifier, record = resolver.resolve(str(pid))
        return record

    def attach_item_ref(self):
        """Attach item reference."""
        item_pid = self.get("item_pid")
        if not item_pid:
            msg = "Missing required field 'item_pid' in loan '{}'"
            raise MissingRequiredParameterError(
                description=msg.format(self["pid"])
            )
        if self.item_ref_builder:
            self["item"] = self.item_ref_builder(self["pid"])

    def update_item_ref(self, new_item_pid):
        """Replace item reference."""
        if not new_item_pid:
            msg = "Missing required arg 'item_pid' when updating loan '{}'"
            raise MissingRequiredParameterError(
                description=msg.format(self["pid"])
            )
        self["item_pid"] = new_item_pid
        self.attach_item_ref()


def is_item_available_for_checkout(item_pid):
    """Return True if the given item is available for loan, False otherwise."""
    config = current_app.config
    cfg_item_can_circulate = config["CIRCULATION_POLICIES"]["checkout"].get(
        "item_can_circulate"
    )
    if not cfg_item_can_circulate(item_pid):
        return False

    search = search_by_pid(
        item_pid=item_pid,
        filter_states=config.get("CIRCULATION_STATES_LOAN_ACTIVE"),
    )
    search_result = search.execute()
    if ES_VERSION[0] >= 7:
        return search_result.hits.total.value == 0
    else:
        return search_result.hits.total == 0


def can_be_requested(loan):
    """Return True if the given record can be requested, False otherwise."""
    config = current_app.config
    cfg_can_be_requested = config["CIRCULATION_POLICIES"]["request"].get(
        "can_be_requested"
    )
    return cfg_can_be_requested(loan)


def get_pending_loans_by_item_pid(item_pid):
    """Return any pending loans for the given item."""
    search = search_by_pid(
        item_pid=item_pid,
        filter_states=current_app.config.get(
            "CIRCULATION_STATES_LOAN_REQUEST"
        ),
    )
    for result in search.scan():
        yield Loan.get_record_by_pid(result["pid"])


def get_pending_loans_by_doc_pid(document_pid):
    """Return any pending loans for the given document."""
    search = search_by_pid(
        document_pid=document_pid,
        filter_states=current_app.config.get(
            "CIRCULATION_STATES_LOAN_REQUEST"
        ),
    )
    for result in search.scan():
        yield Loan.get_record_by_pid(result["pid"])


def get_available_item_by_doc_pid(document_pid):
    """Return an item pid available for this document."""
    for item_pid in get_items_by_doc_pid(document_pid):
        if is_item_available_for_checkout(item_pid):
            return item_pid
    return None


def get_items_by_doc_pid(document_pid):
    """Return a list of item pids for this document."""
    return current_app.config["CIRCULATION_ITEMS_RETRIEVER_FROM_DOCUMENT"](
        document_pid
    )


def get_document_pid_by_item_pid(item_pid):
    """Return the document pid of this item_pid."""
    return current_app.config["CIRCULATION_DOCUMENT_RETRIEVER_FROM_ITEM"](
        item_pid
    )


def get_loan_for_item(item_pid):
    """Return the Loan attached to the given item, if any."""
    loan = None
    if not item_pid:
        return

    search = search_by_pid(
        item_pid=item_pid,
        filter_states=current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"],
    )

    hits = list(search.scan())
    if hits:
        if len(hits) > 1:
            raise MultipleLoansOnItemError(item_pid=item_pid)
        loan = Loan.get_record_by_pid(hits[0]["pid"])
    return loan



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019 CERN.
# Copyright (C) 2018-2019 RERO.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Circulation base transitions."""

import copy
from datetime import datetime

import arrow
from flask import current_app
from invenio_db import db

from ..api import Loan, is_item_available_for_checkout
from ..errors import DocumentDoNotMatchError, DocumentNotAvailableError, \
    InvalidLoanStateError, InvalidPermissionError, ItemNotAvailableError, \
    MissingRequiredParameterError, TransitionConditionsFailedError, \
    TransitionConstraintsViolationError
from ..proxies import current_circulation
from ..signals import loan_state_changed
from ..utils import str2datetime


def ensure_same_patron(f):
    """Validate that the patron PID exists and cannot be changed."""
    def inner(self, loan, **kwargs):
        new_patron_pid = kwargs.get('patron_pid')

        if not current_app.config['CIRCULATION_PATRON_EXISTS'](new_patron_pid):
            msg = "Patron '{0}' not found in the system".format(new_patron_pid)
            raise TransitionConstraintsViolationError(description=msg)

        if loan.get('patron_pid') and new_patron_pid != loan['patron_pid']:
            msg = "Cannot change patron to '{}' while performing an action " \
                  "on this loan".format(new_patron_pid)
            raise TransitionConstraintsViolationError(description=msg)

        return f(self, loan, **kwargs)
    return inner


def ensure_same_document(f):
    """Validate that the document PID exists and cannot be changed."""
    def inner(self, loan, **kwargs):
        new_document_pid = kwargs.get('document_pid')

        if not current_app \
                .config['CIRCULATION_DOCUMENT_EXISTS'](new_document_pid):
            msg = "Document '{0}' not found in the system"\
                .format(new_document_pid)
            raise DocumentNotAvailableError(description=msg)

        if loan.get('document_pid') \
           and new_document_pid != loan['document_pid']:
            msg = "Cannot change document to '{}' while performing an action "\
                  "on this loan".format(new_document_pid)
            raise DocumentDoNotMatchError(description=msg)

        return f(self, loan, **kwargs)
    return inner


def ensure_required_params(f):
    """Decorate to ensure that all required parameters has been passed."""
    def inner(self, loan, **kwargs):
        missing = [p for p in self.REQUIRED_PARAMS if p not in kwargs]
        if missing:
            msg = "Required input parameters are missing '[{}]'" \
                .format(missing)
            raise MissingRequiredParameterError(description=msg)
        if all(param not in kwargs for param in self.PARTIAL_REQUIRED_PARAMS):
            msg = "One of the required parameters '[{}]' is missing." \
                .format(self.PARTIAL_REQUIRED_PARAMS)
            raise MissingRequiredParameterError(description=msg)
        return f(self, loan, **kwargs)
    return inner


def has_permission(f):
    """Decorate to check the transition should be manually triggered."""
    def inner(self, loan, **kwargs):
        if self.permission_factory and not self.permission_factory(loan).can():
            raise InvalidPermissionError(
                permission=self.permission_factory(loan)
            )
        return f(self, loan, **kwargs)
    return inner


def check_trigger(f):
    """Decorate to check the transition should be manually triggered."""
    def inner(self, loan, **kwargs):
        if kwargs.get('trigger', 'next') != self.trigger:
            msg = "The transition with trigger '{}' does not exist."
            raise TransitionConditionsFailedError(
                description=msg.format(self.trigger)
            )
        return f(self, loan, **kwargs)
    return inner


class Transition(object):
    """A transition object that is triggered on conditions."""

    REQUIRED_PARAMS = [
        'transaction_user_pid',
        'patron_pid',
        'transaction_location_pid',
    ]

    PARTIAL_REQUIRED_PARAMS = [
        'item_pid',
        'document_pid'
    ]

    def __init__(self, src, dest, trigger='next', permission_factory=None,
                 **kwargs):
        """Init transition object."""
        self.src = src
        self.dest = dest
        self.trigger = trigger
        self.permission_factory = permission_factory or current_app.config[
            'CIRCULATION_LOAN_TRANSITIONS_DEFAULT_PERMISSION_FACTORY']
        self.validate_transition_states()

    def ensure_item_is_available_for_checkout(self, loan):
        """Validate that an item is available."""
        if 'item_pid' not in loan:
            msg = "Item not set for loan with pid '{}'".format(loan['pid'])
            raise TransitionConstraintsViolationError(description=msg)

        if not current_app.config['CIRCULATION_ITEM_EXISTS'](loan['item_pid']):
            msg = "Item '{0}' not found in catalog".format(loan['item_pid'])
            raise ItemNotAvailableError(description=msg)

        if not is_item_available_for_checkout(loan['item_pid']):
            raise ItemNotAvailableError(
                item_pid=loan['item_pid'], transition=self.dest
            )

    def validate_transition_states(self):
        """Ensure that source and destination states are valid."""
        states = current_app.config['CIRCULATION_LOAN_TRANSITIONS'].keys()
        if not all([self.src in states, self.dest in states]):
            msg = "Source state '{0}' or destination state '{1}' not in [{2}]"\
                .format(self.src, self.dest, states)
            raise InvalidLoanStateError(description=msg)

    def _date_fields2datetime(self, kwargs):
        """Convert any extra kwargs string date to Python datetime."""
        for field in Loan.DATE_FIELDS + Loan.DATETIME_FIELDS:
            if field in kwargs:
                if type(kwargs[field]) is not datetime:
                    kwargs[field] = str2datetime(kwargs[field])

    def before(self, loan, **kwargs):
        """Validate input, evaluate conditions and raise if failed."""
        self.prev_loan = copy.deepcopy(loan)
        loan.update(kwargs)
        loan.setdefault('transaction_date', arrow.utcnow())

    @ensure_same_patron
    @ensure_same_document
    @ensure_required_params
    @has_permission
    @check_trigger
    def execute(self, loan, **kwargs):
        """Execute before actions, transition and after actions."""
        self._date_fields2datetime(kwargs)
        loan.date_fields2datetime()

        self.before(loan, **kwargs)
        loan['state'] = self.dest
        self.after(loan)

    def after(self, loan):
        """Commit record and index."""
        self.prev_loan.date_fields2str()
        loan.date_fields2str()

        loan.commit()
        db.session.commit()
        current_circulation.loan_indexer.index(loan)

        loan_state_changed.send(
            self, prev_loan=self.prev_loan, loan=loan, trigger=self.trigger)
